// *************************************************************************
// * autor:   Marek Sopko                                                  *
// * kontakt: ... Trencin, Slovenska republika                             *
// *          http:   www.avatar.host.sk                                   *
// *          e-mail: avatar&host.sk                                       *
// *          mobil:  +421 907 952 453                                     *
// *                                                                       *
// * popis aplikacie: 3D Engine (vyuziva OpenGL API)                       *
// *                  vykreslovaci stroj, ktory pre pouzivatela vytvara    *
// *                  virtualny trojrozmerny svet, v ktorom sa moze volne  *
// *                  pohybovat a vykonavat vopred definovanu cinnost      *
// *************************************************************************

// obsahom mEngine.cpp su nasledovne funkcie:
// hlavna vykonavacia funkcia aplikacie - main()
// vytvorenie SDL okna
// obsluha mysi
// obsluha klavesnice

// hlavickove subory *******************************************************
// GL kniznica, pouzije sa subor /usr/lib/libGL.so, resp. subor
// patriaci konkretnej grafickej karte ako napr. /usr/lib/nvidia/libGL.so
// v pripade pouzitia GLU kniznice sa pouzije subor /usr/lib/libGLU.so
// v pripade pouzitia GLEXT kniznice sa pouzije subor suvisiaci s GPU, napr.
// /usr/lib/xorg/modules/extensions/nvidia/libglx.so
//
// pouzitie jednotlivych kniznic je zavisle na pouzivanom grafickom hardveri
// a tiez implementiacii OpenGL, napr. pomocou Mesa kniznice
#include <GL/gl.h>
// funkcie SDL kniznice (pouzije sa subor /usr/lib/libSDL.so)
#include <SDL/SDL.h>

// vlastny .h subor
//#include "mEngine.h"

// hlavny interface aplikacie
#include "mInterface.h"

/***************************************************************************
 * deklaracie a definicie premennych a konstant                            *
 ***************************************************************************/
// sirka, vyska a farebna hlbka (pocet bitov na 1 pixel) SDL okna
#define SCR_WIDTH  1080
#define SCR_HEIGHT 729
#define SCR_BPP    32
// pozn.: v X11, VidMode nemeni rozlisenie obrazovky, vo Win32,
// ChangeDisplaySettings moze zmenit BPP

// ukoncenie hlavneho cyklu
int isDone=0;
// SDL okno je/nie je aktivne
int isActive=1;
// polovica sirky a vysky okna
int scr_width2=0;
int scr_height2=0;
// meno aplikacie
const char app_title[64]="Engine v1.0 (OpenGL API)";

// premenne potrebne pre SDL
const SDL_VideoInfo *videoInfo=NULL; // informacie o obrazovke
int videoFlags=0;                    // znacky pre SDL_SetVideoMode
SDL_Event event;                     // obsluha udalosti
SDL_Surface *surface=NULL;           // SDL kresliaca plocha
Uint8 *keystate=NULL;                // SDL pole stlacanych klaves

// pocas startu aplikacie sa nebude brat do uvahy poloha mysi
int go_mouse=0;

// integracia OpenGL
mInterface *Interface=NULL; // hlavny interface aplikacie

// identifikator Avatara - pozorovatela
unsigned int Aid=0;

/***************************************************************************
 * definicie funkcii aplikacie                                             *
 ***************************************************************************/
// uvolnenie alokovanych zdrojov a obnovenie povodnych nastaveni obrazovky *
void APP_quit(int returnCode)
{
 if (Interface!=NULL) delete Interface; // zrusenie instancie triedy mInterface
 Interface=NULL;

 SDL_Quit(); // zrusenie SDL okna

 exit(returnCode); // ukoncenie aplikacie
}

// inicializacia SDL video subsystemu **************************************
int SDL_initialize(void)
{
 // inicializacia SDL video subsystemu
 if (SDL_Init(SDL_INIT_VIDEO)<0) {
  fprintf(stderr,"Chyba inicializacie SDL video subsystemu: %s\n",SDL_GetError());
  APP_quit(1);
 }

 // ziskanie informacii o SDL video subsysteme
 videoInfo=SDL_GetVideoInfo();
 if (!videoInfo) {
  fprintf(stderr,"Chyba ziskania informacii o SDL video subsysteme: %s\n",SDL_GetError());
  APP_quit(1);
 }

 // nastavenie atributov SDL kresliacej plochy
 // 8 bitov pre cervenu, zelenu a modru farbu
 // 8 bitov pre alfa kanal
 // 16 bitovy zasobnik hlbky
 // 1 pre pouzitie dvojiteho zasobnika pre prekreslovanie
 // 1 bitovy stencil zasobnik
 SDL_GL_SetAttribute(SDL_GL_RED_SIZE,    8);
 SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE,  8);
 SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE,   8);
 SDL_GL_SetAttribute(SDL_GL_ALPHA_SIZE,  8);
 SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE,  16);
 SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER,1);
 //SDL_GL_SetAttribute(SDL_GL_STENCIL_SIZE,1);
 // pozn.: SDL_DOUBLEBUFFER sa pouziva v SDL_GL_SetAttribute,
 // v SDL_SetVideoMode to ovplivni iba 2D prekreslovanie

 // znacky pre SDL_SetVideoMode
 videoFlags =SDL_OPENGL;          // zapne OpenGL v SDL
 videoFlags|=SDL_GL_DOUBLEBUFFER; // alokuje dvojity zasobnik pre prekreslovanie
 videoFlags|=SDL_HWPALETTE;       // ulozi paletu do HW
 //videoFlags|=SDL_RESIZABLE;       // umozni menit velkost SDL okna

 // kontrola ci mozno SDL kresliacu plochu umiestnit do HW
 if (videoInfo->hw_available) videoFlags|=SDL_HWSURFACE;
 else videoFlags|=SDL_SWSURFACE;

 // kontrola, ci mozno pracovat s HW blit
 // blitting - kopia casti zdrojovej textury do cielovej textury
 if (videoInfo->blit_hw) videoFlags|=SDL_HWACCEL;

 // nastavenie SDL kresliacej plochy
 surface=SDL_SetVideoMode(SCR_WIDTH,SCR_HEIGHT,SCR_BPP,videoFlags);
 if (!surface) {
  fprintf(stderr,"Chyba nastavenia SDL kresliacej plochy: %s\n",SDL_GetError());
  APP_quit(1);
 }

 // nastavenie nazvu okna a ikony
 SDL_WM_SetCaption(app_title,app_title);
 // schovanie kurzoru mysi
 SDL_ShowCursor(SDL_DISABLE);

 return(0);
}

// modifikacia nastaveni kresliacej plochy po zmene velkosti okna **********
int WND_resize(int width,int height)
{
 // zmena SDL kresliacej plochy (okna)
 surface=SDL_SetVideoMode(width,height,SCR_BPP,videoFlags);
 if (!surface) {
  fprintf(stderr,"Chyba nastavenia SDL kresliacej plochy po zmene velkosti okna: %s\n",SDL_GetError());
  APP_quit(1);
 }

 // INFO: doslo k zmene velkosti okna
 //fprintf(stderr,"WND_resize ... ok\n");

 // reinicializacia nastaveni Avatara po zmene velkosti okna
 if (Interface!=NULL) Interface->Avatar[Aid].cam_reinit((GLfloat) width,(GLfloat) height);

 return(0);
}

// obhospodarenie vstupu z mysi ********************************************
void APP_mouse(int button,int state,int x,int y)
{
 // stlacenie laveho a praveho tlacidla mysi
 Interface->Avatar[Aid].isFire1=SDL_GetMouseState(NULL,NULL)&SDL_BUTTON(SDL_BUTTON_LEFT);
 Interface->Avatar[Aid].isFire2=SDL_GetMouseState(NULL,NULL)&SDL_BUTTON(SDL_BUTTON_RIGHT);

 // go_mouse=1 az po prvom nastaveni kurzoru na stred okna,
 // predijde sa tak nechcenej rotacii kamery po starte aplikacie
 x=(x-scr_width2)*go_mouse; y=(y-scr_height2)*go_mouse;

 // rotacia kamery na zakladne pohybu mysi
 Interface->Avatar[Aid].cam_rotate_xy((GLfloat) y,(GLfloat) x);
}

// obhospodarenie vstupu z klavesnice **************************************
void APP_keyboard(void)
{
 keystate=SDL_GetKeyState(NULL); // nacitanie stlacenych klaves

 isDone=(int) keystate[SDLK_ESCAPE];                              // ESC

 Interface->Avatar[Aid].goForward=(GLfloat) keystate[SDLK_UP];    // sipka hore
 Interface->Avatar[Aid].goBack=   (GLfloat) keystate[SDLK_DOWN];  // sipka dole
 Interface->Avatar[Aid].goRight=  (GLfloat) keystate[SDLK_RIGHT]; // sipka doprava
 Interface->Avatar[Aid].goLeft=   (GLfloat) keystate[SDLK_LEFT];  // sipka dolava
 Interface->Avatar[Aid].goUp=     (GLfloat) keystate[SDLK_a];     // posun kamery hore
 Interface->Avatar[Aid].goDown=   (GLfloat) keystate[SDLK_z];     // posun kamery dole
 // skok
 Interface->Avatar[Aid].goJump=keystate[SDLK_j];
 // chodza alebo beh
 Interface->Avatar[Aid].goFast=(keystate[SDLK_RSHIFT]||keystate[SDLK_LSHIFT]);

 // rotacia kamery okolo doprednej osi
 Interface->Avatar[Aid].cam_rotate_zl((GLfloat) keystate[SDLK_q]);
 Interface->Avatar[Aid].cam_rotate_zr((GLfloat) keystate[SDLK_w]);

 Interface->pause_motion=(GLubyte) keystate[SDLK_p]; // zap/vyp casovo zavisleho pohybu objektov
 Interface->help_onoff=  (GLubyte) keystate[SDLK_h]; // zap/vyp vypisu helpu

 if (keystate[SDLK_l]) Interface->change_polygon_mode(); // vyplnene trojuholniky/iba hrany
 if (keystate[SDLK_d]) Interface->change_LOD();          // zmena urovne detailov

 if (keystate[SDLK_r]) Interface->Avatar[Aid].cam_reset(); // reset pohladu kamery

 // posun kamery na zaklade vstupu z klavesnice
 Interface->Avatar[Aid].cam_move();
}

// funkcia pre kreslenie OpenGL sceny **************************************
int OGL_SDL_draw_scene(void)
{
 // vykreslenie OpenGL sceny
 Interface->ogl_Draw();

 // vymena zobrazovacich zasobnikov, zobrazenie sceny
 SDL_GL_SwapBuffers();

 return(0);
}

// hlavna vykonavacia funkcia aplikacie ************************************
int main(int argc,char **argv)
{
 scr_width2=SCR_WIDTH/2;   // vypocet polovice sirky okna
 scr_height2=SCR_HEIGHT/2; // vypocet polovice vysky okna

 // inicializacia SDL
 SDL_initialize();

 // tvorba instancie triedy mInterface
 Interface=new mInterface();
 if (Interface==NULL) {
  fprintf(stderr,"Chyba tvorby instancie triedy mInterface!\n");
  APP_quit(1);
 }

 // inicializacia OpenGL 0=OK, 1=ERROR
 if (Interface->ogl_Initialize()) APP_quit(1);

 // upravenie velkosti SDL okna
 WND_resize(SCR_WIDTH,SCR_HEIGHT);

 // zistenie rozmerov okna - potrebne pre operacie v triede Interface
 glGetIntegerv(GL_VIEWPORT,Interface->ViewPort);

 // reset kamery z dovodu spravneho nastartovania casovaca
 Interface->Avatar[Aid].cam_resettime();

 // hlavny programovy cyklus
 while (!isDone) {
  // ukladanie a spracovavanie udalosti
  while (SDL_PollEvent(&event)) {
   switch (event.type) {
    case SDL_ACTIVEEVENT: // kreslenie iba ak je okno aktivne
     if (event.active.gain) { // ak dojde k aktivacii okna, restartuje sa casovac
      Interface->Avatar[Aid].cam_resettime();
      isActive=1;
     } else isActive=0;       // okno nie je aktivne
     break;
    case SDL_VIDEORESIZE: // zmena velkosti okna
     WND_resize(event.resize.w,event.resize.h);
     break;
    case SDL_QUIT:        // zrusenie SDL okna
     isDone=1;
     break;
    case SDL_MOUSEMOTION: // obsluha mysi
     APP_mouse(event.button.button,event.button.state,event.button.x,event.button.y);
     break;
    default: break;
   }
  }
  // obsluha klavesnice
  APP_keyboard();
  // update parametrov kamery
  Interface->Avatar[Aid].cam_update();
  // casovo zavisla uprava objektov, resp. sceny
  Interface->ogl_Iddle();

  // umiestnenie kurzoru mysi na stred okna
  SDL_WarpMouse(SCR_WIDTH/2,SCR_HEIGHT/2);

  // od tohto okamihu moze byt kamera rotovana na zaklade polohy mysi
  go_mouse=1;

  // kreslenie sceny
  if (isActive) OGL_SDL_draw_scene();
 }

 // ukoncenie aplikacie
 APP_quit(0);

 return(0);
}
